;##
;# Associate registry entry for music files for snackAmp
;# Note: Windows only, will not be called on Unix/Mac
;# Proudly stolen from WaveSurfer a very cool sound editor
;# Tom Wilkason
;#
package require registry
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/fileAssociation.tcl,v 1.5 2002/10/14 14:32:50 wilkason Exp $
}
###############################################################################
# Function   : isRegisteredMF
# Description: Determine if a file is registered to SnackAmp
###############################################################################
proc isRegisteredMF {{ftype .mp3} _var} {
   upvar $_var var
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      set var 0
      return 0
   }
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore\\$class\\DefaultIcon"
   if {[catch {registry get $resKey ""} defIcon]} {
      set var 0
      return 0
   }
   set var 1
   return 1
}
###############################################################################
# Function   : unRegisterMF
# Description: UnRegister some file type on windows for launching snackamp
###############################################################################
proc unRegisterMF {{ftype .mp3}} {
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      puts "Not registered to snackAmp"
      return
   }
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to restpre $ftype associations to their previous settings?"]!="yes"} {
      return
   } ;# end if
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   ;# Restore single entries for
   #HKEY_CLASSES_ROOT .mp3
   #HKEY_CLASSES_ROOT mp3file
   ;# Remove snAmp.append handlers
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.append
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.play
   ;# Restore old Icon
   if {![catch {registry get $resKey\\$class\\DefaultIcon ""} file]} {
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" $file
      registry delete $resKey\\$class\\DefaultIcon
   }
   if {![catch {registry get $resKey\\$class\\shell ""} value]} {
      registry set HKEY_CLASSES_ROOT\\$class\\shell "" $value
      registry delete $resKey\\$class\\shell
   }
}
###############################################################################
# Function   : registerMF
# Description: Register some file type on windows for launching snackamp
###############################################################################
proc registerMF {{ftype .mp3}} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to associate $ftype files with SnackAmp"]!="yes"} {
      return
   } ;# end if

   package require registry
   set conf($ftype,default) 1
   set conf($ftype,dde) 1
   set conf($ftype,assoc) $ftype
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   foreach key [array names conf *,assoc] {
      set ext [lindex [split $key ,] 0]
      ;# Set the class name of the file type
      if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
         set class "[string trimleft $ext .]file"
         registry set [join [list HKEY_CLASSES_ROOT $ext] \\] "" $class
         registry set [join [list HKEY_CLASSES_ROOT $class] \\] "" "[string toupper [string trimleft $ext .]] sound format"
      }
      ;# Save the existing icon
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\DefaultIcon ""} file]} {
         registry set $resKey\\$class\\DefaultIcon "" $file
      }
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\shell ""} value]} {
         registry set $resKey\\$class\\shell "" $value
      }
      ;# Register the default icon for this class to SnackAmp (only gets icon if wrapped)
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" "$::TA(appname)"
      set base HKEY_CLASSES_ROOT\\$class\\shell
      ;##
      ;# Add menu handlers (These can just be removed later)
      ;#
      ;# Add son(s) to playlist
      if {$conf($ext,default)} {
         registry set $base "" snAmp.append
      }
      registry set $base\\snAmp.append "" "&Append to SnackAmp Playlist"
      registry set $base\\snAmp.append\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.append\\ddeexec "" "Append {%1}"
         registry set $base\\snAmp.append\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.append\\ddeexec\\Topic "" snackAmp
      }
      ;# Play the file
      if {$conf($ext,default)} {
         registry set $base "" snAmp.play
      }
      registry set $base\\snAmp.play "" "&Play using SnackAmp"
      registry set $base\\snAmp.play\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.play\\ddeexec "" "PlayFile {%1}"
         registry set $base\\snAmp.play\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.play\\ddeexec\\Topic "" snackAmp
      }
   }
}
###############################################################################
# Description: During initialization, save the name of the executable
###############################################################################
if {[info exists ::wrap]} {
   set TA(appname) "\"[file nativename [info nameofexecutable]]\""
} else {
   set TA(appname) "\"[file nativename [info nameofexecutable]]\" \"[info script]\""
}
package provide fileAssociation 1.0
