#!/usr/bin/wish
#
# Copyright (C) 1997,1998 D. Richard Hipp
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
#
# Author contact information:
#   drh@acm.org
#   http://www.hwaci.com/drh/
#
# $Revision: 1.12 $
# Modified by Tom Wilkason to add checkboxes for each node if they are linked to a variable
# 1/9/2001
# Hack Note: By default a closed folder is the icon, and when selected it turns into an open folder
# This could be changed to pass in openimage, closeimage but for this purpose it saves memory.
# For generalization it should be done this way however.
#
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saTree.tcl,v 1.12 2002/10/14 14:32:33 wilkason Exp $
}
option add *highlightThickness 0

switch $tcl_platform(platform) {
   unix {
      set Tree(font) \
            -adobe-helvetica-medium-r-normal-*-11-80-100-100-p-56-iso8859-1
   }
   windows {
      set Tree(font) \
            -adobe-helvetica-medium-r-normal-*-14-100-100-100-p-76-iso8859-1
   }
   set Tree(fg) black
   set Tree(bg) white
}

#
# Create a new tree widget.  $args become the configuration arguments to
# the canvas widget from which the tree is constructed.
#
proc Tree:create {w args} {
   global Tree
   eval canvas $w $args
   bind $w <Destroy> [list Tree:delitem $w /]
   ;# cb to change outline
   bind [winfo parent $w] <FocusIn> [list Tree:FocusChange $w 1]
   bind [winfo parent $w] <FocusOut> [list Tree:FocusChange $w 0]
   Tree:dfltconfig $w /
   Tree:buildwhenidle $w
   set Tree($w:selection) {}
   set Tree($w:selidx) {}
   bind $w <Configure> [list Tree:canvasResize %W %w %h]
}

;##
;# Called on a canvas resize or configure event, need to
;# The canvas size and the scroll region
proc Tree:canvasResize {win W H} {
   set bbox [$win bbox all]
   $win configure -scrollregion $bbox -yscrollincrement 0.1i
}
# Initialize a element of the tree.
# Internal use only
#
proc Tree:dfltconfig {w v} {
   global Tree
   set Tree($w:$v:children) {}
   set Tree($w:$v:open) 0
   #set Tree($w:$v:icon) Tree:Folder   ;# Default use tree
   set Tree($w:$v:tags) {}
}

#
# Pass configuration options to the tree widget
#
proc Tree:config {w args} {
   eval $w config $args
}

#
# Insert a new element $v into the tree $w.
#
proc Tree:newitem {w v args} {
   global Tree nodeState
   ;# If windows path or unc, strip drive letters / double slashes to make a Unix style root
   set v [string map {// / : "_Drive"} $v]
   ;# Force this to be off the root in case user forget leading /
   if {[string index $v 0] != "/"} {
      set v "/$v"
   }
   set dir [file dirname $v]
   set n [file tail $v]
   # puts "$v - $dir - $n"
   ;# If this is a root
   if {![info exists Tree($w:$dir:open)]} {
      # puts "parent item \"$dir\" for $v (tail $n) is missing - creating one"
      eval Tree:newitem $w $dir $args
   }
   set i [lsearch -exact $Tree($w:$dir:children) $n]
   if {$i>=0} {
      error "item \"$v\" already exists"
      ;# Need to ignore this
   }
   lappend Tree($w:$dir:children) $n
   set Tree($w:$dir:children) [lsort -dictionary $Tree($w:$dir:children)]
   Tree:dfltconfig $w $v
   foreach {op arg} $args {
      switch -exact -- $op {
         -image {set Tree($w:$v:icon) $arg}
         -tags {set Tree($w:$v:tags) $arg}
         -variable {
            upvar #0 $arg myVar
            set Tree($w:$v:variable) $arg
            set nodeState($v) $myVar
            # puts "$v -> $myVar"
         }
      }
   }
   Tree:buildwhenidle $w
}
#
# Delete element $v from the tree $w.  If $v is /, then the widget is
# deleted.
#
proc Tree:delitem {w v} {
   global Tree
   if {![info exists Tree($w:$v:open)]} return
   if {[string compare $v /]==0} {
      # delete the whole widget
      catch {destroy $w}
      foreach t [array names Tree $w:*] {
         unset Tree($t)
      }
      return
   }
   foreach c $Tree($w:$v:children) {
      catch {Tree:delitem $w [file join $v $c]}
   }
   unset Tree($w:$v:open)
   unset Tree($w:$v:children)
   unset Tree($w:$v:icon)
   set dir [file dirname $v]
   set n [file tail $v]
   set i [lsearch -exact $Tree($w:$dir:children) $n]
   if {$i>=0} {
      set Tree($w:$dir:children) [lreplace $Tree($w:$dir:children) $i $i]
   }
   Tree:buildwhenidle $w
}

#
# Change the selection to the indicated item
#
proc Tree:setselection {w v} {
   global Tree
   set Tree($w:selection) $v
   Tree:drawselection $w
}

#
# Retrieve the current selection
#
proc Tree:getselection w {
   global Tree
   return $Tree($w:selection)
}

#
# Retrieve the current selection tag
#
proc Tree:activetag w {
   global Tree
   return $Tree($w:activeTag)
}
#
# Bitmaps used to show which parts of the tree can be opened.
#
set maskdata "#define solid_width 9\n#define solid_height 9"
append maskdata {
   static unsigned char solid_bits[] = {
      0xff, 0x01, 0xff, 0x01, 0xff, 0x01, 0xff, 0x01, 0xff, 0x01, 0xff, 0x01,
      0xff, 0x01, 0xff, 0x01, 0xff, 0x01
   };
}
set data "#define open_width 9\n#define open_height 9"
append data {
   static unsigned char open_bits[] = {
      0xff, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x7d, 0x01, 0x01, 0x01,
      0x01, 0x01, 0x01, 0x01, 0xff, 0x01
   };
}
image create bitmap Tree:openbm -data $data -maskdata $maskdata -foreground black -background white
#image create bitmap Tree:unselectbm -data $maskdata -maskdata $maskdata -foreground white -background blue
set data "#define closed_width 9\n#define closed_height 9"
append data {
   static unsigned char closed_bits[] = {
      0xff, 0x01, 0x01, 0x01, 0x11, 0x01, 0x11, 0x01, 0x7d, 0x01, 0x11, 0x01,
      0x11, 0x01, 0x01, 0x01, 0xff, 0x01
   };
}
image create bitmap Tree:closedbm -data $data -maskdata $maskdata -foreground black -background white
#image create bitmap Tree:selectbm -data $maskdata -maskdata $maskdata -foreground blue -background white
# Check
image create photo Tree:selectbm -data {
   R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
   AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
   AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
   ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
   AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
   QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
   QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
   QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
   QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
   gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
   gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
   gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
   gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
   gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
   wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
   wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
   wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
   AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAg8AEkJHEiQ4L+D
   CBP+E6iwIcOG/wAcfKgQgMSFpA5ejGhxYkaOHDc+tEgyIcWSJj9qdKgSIkaX
   CAvKHBgQADs=
}
image create photo Tree:unselectbm -data {
   R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
   AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
   AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
   ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
   AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
   QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
   QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
   QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
   QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
   gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
   gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
   gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
   gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
   gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
   wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
   wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
   wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
   AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAgtAEkJHEiQ4L+D
   CBP+E6iwIcOGCR9CPChxYkWIFx2Smogwo0KPETdyXFiwpMCAADs=
}
image create photo Tree:parselectbm -data {
   R0lGODdhDQANAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgMDAwMDcwKbK8EAg
   AGAgAIAgAKAgAMAgAOAgAABAACBAAEBAAGBAAIBAAKBAAMBAAOBAAABgACBg
   AEBgAGBgAIBgAKBgAMBgAOBgAACAACCAAECAAGCAAICAAKCAAMCAAOCAAACg
   ACCgAECgAGCgAICgAKCgAMCgAOCgAADAACDAAEDAAGDAAIDAAKDAAMDAAODA
   AADgACDgAEDgAGDgAIDgAKDgAMDgAODgAAAAQCAAQEAAQGAAQIAAQKAAQMAA
   QOAAQAAgQCAgQEAgQGAgQIAgQKAgQMAgQOAgQABAQCBAQEBAQGBAQIBAQKBA
   QMBAQOBAQABgQCBgQEBgQGBgQIBgQKBgQMBgQOBgQACAQCCAQECAQGCAQICA
   QKCAQMCAQOCAQACgQCCgQECgQGCgQICgQKCgQMCgQOCgQADAQCDAQEDAQGDA
   QIDAQKDAQMDAQODAQADgQCDgQEDgQGDgQIDgQKDgQMDgQODgQAAAgCAAgEAA
   gGAAgIAAgKAAgMAAgOAAgAAggCAggEAggGAggIAggKAggMAggOAggABAgCBA
   gEBAgGBAgIBAgKBAgMBAgOBAgABggCBggEBggGBggIBggKBggMBggOBggACA
   gCCAgECAgGCAgICAgKCAgMCAgOCAgACggCCggECggGCggICggKCggMCggOCg
   gADAgCDAgEDAgGDAgIDAgKDAgMDAgODAgADggCDggEDggGDggIDggKDggMDg
   gODggAAAwCAAwEAAwGAAwIAAwKAAwMAAwOAAwAAgwCAgwEAgwGAgwIAgwKAg
   wMAgwOAgwABAwCBAwEBAwGBAwIBAwKBAwMBAwOBAwABgwCBgwEBgwGBgwIBg
   wKBgwMBgwOBgwACAwCCAwECAwGCAwICAwKCAwMCAwOCAwACgwCCgwECgwGCg
   wICgwKCgwMCgwOCgwADAwCDAwEDAwGDAwIDAwKDAwP/78KCgpICAgP8AAAD/
   AP//AAAA//8A/wD//////yH5BAAAAAAALAAAAAANAA0AAAg8APMJHEiQ4L+D
   CBP+E6iwIcOG/w4cfKjwgMSF+Q5ejGhxYkaOHDc+tEgyIcWSJj9qdKgSIkaX
   CAvKHBgQADs=
}
image create photo Tree:ROM.open -data {
   R0lGODlhEAAQAIQAAPwCBFxaXISChERCBMTCxKSipAQCBMTCBPzCxPz+xPze
   rMT+xDQyNMTC/ASCBPyqXPz+BARCBAT+BAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAWIICCKgWAG
   YyoKA+EWhGGogOAaB6Lr8ygUCYWCkEMoEovBSHZUFI8LGYloODJkhWijMbMd
   iAwBQEbYxmpExFlkgN3QrliXAcvNAuVt4coYKtQ9BlwLCQQFMAlVSmwEUX8I
   B1UJiQAObEE6BA9VEDMOEUsDDzoHkZ4LoCkyqyIOCxI9NCOfETJ+IQAh/mhD
   cmVhdGVkIGJ5IEJNUFRvR0lGIFBybyB2ZXJzaW9uIDIuNQ0KqSBEZXZlbENv
   ciAxOTk3LDE5OTguIEFsbCByaWdodHMgcmVzZXJ2ZWQuDQpodHRwOi8vd3d3
   LmRldmVsY29yLmNvbQA7
}
image create photo Tree:ROM.close -data {
   R0lGODlhEAAQAIUAAPwCBFxaXISChNTS1MzKzMTGxMy+vNSytAQCBOzq7OTi
   5Nza3Oy6vPy2tOTStPz6/Ozu7Nze3MTKxOS2tNzmvLT2tDw+BPTy9MzOzLS2
   tPy+tLz+tLz2tLSynLzqvLzavLTOtNTW1KSipPz+/MTCxLy6vDQyNOTm5AAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAap
   QIBQGBAYA8Ok0DggFAwHBEIJECQUi8GT0XBMh9YHJDKQTLqUigWsuDwSmMxB
   Q9lwvoEBQpEoBBAWHRseHyBTAgMhCxgWIiMWFiASJCSHEgQSGQgjmwglGKCH
   BQUSICYWIyIWAwshIVMBIB8eFSQmCAEFCREIekIIFRsUGgcZEgkPFydrvxUU
   Dg0GBSEREGJfv9AMEySsJxDYQ4DcEhh64UpS6lTs7QB+QQAh/mhDcmVhdGVk
   IGJ5IEJNUFRvR0lGIFBybyB2ZXJzaW9uIDIuNQ0KqSBEZXZlbENvciAxOTk3
   LDE5OTguIEFsbCByaWdodHMgcmVzZXJ2ZWQuDQpodHRwOi8vd3d3LmRldmVs
   Y29yLmNvbQA7
}
;# TFW: Hack, Tree:closeFolder/Tree:openFolder need to be
;# passed in not hardcoded here
image create photo Tree:Folder.close -data {
   R0lGODlhEAAQAIYAAPwCBAQCBExKTBwWHMzKzOzq7ERCRExGTCwqLARqnAQ+
   ZHR2dKyqrNTOzHx2fCQiJMTi9NTu9HzC3AxmnAQ+XPTm7Dy67DymzITC3IzG
   5AxypHRydKymrMzOzOzu7BweHByy9AyGtFyy1IzG3NTu/ARupFRSVByazBR6
   rAyGvFyuzJTK3MTm9BR+tAxWhHS61MTi7Pz+/IymvCxulBRelAx2rHS63Pz6
   /PTy9PTu9Nza3ISitBRupFSixNTS1CxqnDQyNMzGzOTi5MTCxMTGxGxubGxq
   bLy2vLSutGRiZLy6vLSytKyurDQuNFxaXKSipDw6PAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAA
   LAAAAAAQABAAAAfDgACCAAECg4eIAAMEBQYHCImDBgkKCwwNBQIBBw4Bhw8Q
   ERITFJYEFQUFnoIPFhcYoRkaFBscHR4Ggh8gIRciEiMQJBkltCa6JyUoKSkX
   KhIrLCQYuQAPLS4TEyUhKb0qLzDVAjEFMjMuNBMoNcw21QY3ODkFOjs82RM1
   PfDzFRU3fOggcM7Fj2pAgggRokOHDx9DhhAZUqQaISBGhjwMEvEIkiIHEgUA
   kgSJkiNLmFSMJChAEydPGBSBwvJQgAc0/QQCACH+aENyZWF0ZWQgYnkgQk1Q
   VG9HSUYgUHJvIHZlcnNpb24gMi41DQqpIERldmVsQ29yIDE5OTcsMTk5OC4g
   QWxsIHJpZ2h0cyByZXNlcnZlZC4NCmh0dHA6Ly93d3cuZGV2ZWxjb3IuY29t
   ADs=
}
image create photo Tree:Folder.open -data {
   R0lGODlhEAAQAIYAAPwCBAQCBExKTBQWFOzi1Ozq7ERCRCwqLPz+/PT29Ozu
   7OTm5FRSVHRydIR+fISCfMTCvAQ6XARqnJSKfIx6XPz6/MzKxJTa9Mzq9JzO
   5PTy7OzizJSOhIyCdOTi5Dy65FTC7HS2zMzm7OTSvNTCnIRyVNza3Dw+PASq
   5BSGrFyqzMyyjMzOzAR+zBRejBxqnBx+rHRmTPTy9IyqvDRylFxaXNze3DRu
   jAQ2VLSyrDQ2NNTW1NTS1AQ6VJyenGxqbMTGxLy6vGRiZKyurKyqrKSmpDw6
   PDw6NAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAA
   LAAAAAAQABAAAAfCgACCAAECg4eIAAMEBQYCB4mHAQgJCgsLDAEGDQGIkw4P
   BQkJBYwQnRESEREIoRMUE6IVChYGERcYGaoRGhsbHBQdHgu2HyAhGSK6qxsj
   JCUmJwARKCkpKsjKqislLNIRLS4vLykw2MkRMRAGhDIJMzTiLzDXETUQ0gAG
   CgU2HjM35N3AkYMdAB0EbCjcwcPCDBguevjIR0jHDwgWLACBECRIBB8GJekQ
   MiRIjhxEIlBMFOBADR9FIhiJ5OnAEQB+AgEAIf5oQ3JlYXRlZCBieSBCTVBU
   b0dJRiBQcm8gdmVyc2lvbiAyLjUNCqkgRGV2ZWxDb3IgMTk5NywxOTk4LiBB
   bGwgcmlnaHRzIHJlc2VydmVkLg0KaHR0cDovL3d3dy5kZXZlbGNvci5jb20A
   Ow==
}
array set ::TreeImages {
   1  Tree:selectbm
   true  Tree:selectbm
   True  Tree:selectbm
   0  Tree:unselectbm
   false  Tree:unselectbm
   False  Tree:unselectbm
   2  Tree:parselectbm
}

# Internal use only.
# Draw the tree on the canvas
proc Tree:build {w} {
   global Tree
   $w delete all
   ;# Remove each of the old tag entries
   foreach tag [array names Tree *:tag:*] {
      unset Tree($tag)
   } ;# end foreach
   catch {unset Tree($w:buildpending)}
   set Tree($w:y) 30
   Tree:buildlayer $w / 10      ;# This builds the root
   $w config -scrollregion [$w bbox all]
   Tree:drawselection $w
}

# Internal use only.
# Build a single layer of the tree on the canvas.  Indent by $in pixels
proc Tree:buildlayer {w v in} {
   global Tree
   # puts "Tree:buildlayer $w $v $in"
   set deltaHeight [font metrics $Tree(font) -linespace]  ;# Node spacing is line spacing
   set deltaHeight [expr {$deltaHeight < 15?15:$deltaHeight}]  ;# Clamp to min size
   set y $Tree($w:y)
   set start [expr {$y-10}]
   foreach c $Tree($w:$v:children) {
      set y $Tree($w:y)
      incr Tree($w:y) $deltaHeight ;#17
      $w create line $in $y [expr {$in+12}] $y -fill gray ;#50
      set token [file join $v $c]
      # token is the folder name
      set icon $Tree($w:$token:icon)
      set taglist x
      foreach tag $Tree($w:$token:tags) {
         lappend taglist $tag
      }
      set x [expr {$in+10}]  ;# how far over to place the icon
      ;##
      ;# Create the user specified icon
      ;#
      if {[string length $icon] > 0} {
         set k [$w create image $x $y -image $icon.close -anchor w -tags $taglist]
         incr x 20
         set Tree($w:tag:$k) $token
      }
      ;# Toggle button if user wants it (linked to a variable)
      if {[info exists Tree($w:$token:variable)]} {
         upvar #0 $Tree($w:$token:variable) myVar
         incr x 5
         set cb [$w create image $x $y -image $::TreeImages($myVar)]
         ;# On a click, need to change indicator and set a global
         ;# vx = base dir, c = file name

         ;# TFW: Note we are setting ::TA(APLDirty) here, non-generic!
         $w bind $cb <1> "[list Tree:toggleNode $w $cb $token];set ::TA(APLDirty) 1"
         incr x 10
      }
      set j [$w create text $x $y -text $c -font $Tree(font) -fill $Tree(fg) -activefill $Tree(fg) \
         -anchor w -tags $taglist]
      set Tree($w:tag:$j) $token
      set Tree($w:$token:tag) $j
      ;# Create the +/- images to click on to expand the tree
      if {[string length $Tree($w:$token:children)]} {
         if {$Tree($w:$token:open)} {
            ;# - size box
            set j [$w create image $in $y -image Tree:openbm]
            $w bind $j <1> "[list set Tree($w:$token:open) 0];[list Tree:build $w]"
            Tree:buildlayer $w $token [expr {$in+18}]
         } else {
            ;# + sign box
            set j [$w create image $in $y -image Tree:closedbm]
            $w bind $j <1> "[list set Tree($w:$token:open) 1];[list Tree:build $w]"
         }
      }
   }
   set j [$w create line $in $start $in [expr {$y+1}] -fill gray]
   $w lower $j
}
###############################################################################
# Function   : Tree:toggleNode
# Description: Bound to checkbox at node
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc Tree:toggleNode {can tag node} {
   global Tree nodeState
   set index $can:$node
   upvar $Tree($index:variable) linkedVariable
   set nodeState($node) [expr {!$nodeState($node)}]
   if {$nodeState($node)} {
      $can itemconfigure $tag -image Tree:selectbm
      ;##
      ;# TFW: Need to set the parent tag to a 2 IFF
      ;#      it is currently 0. tag is an integer
      ;# nodeState is indexed by folder with 0/1/2 values
      ;# Tree is a complex data structure.
      ;#
   } else {
      $can itemconfigure $tag -image Tree:unselectbm
   }
   set linkedVariable $nodeState($node)
}
proc Tree:toggleExpand {w v {recurse 0}} {
   global Tree
   set Tree(depth) 0
   if {[info exists Tree($w:$v:open)] && $Tree($w:$v:open)==1} {
      Tree:close $w $v $recurse
   } else {
      Tree:open  $w $v $recurse
   }
}
# Open a branch of a tree
#
proc Tree:open {w v {recurse 0}} {
   global Tree
   # puts "Tree:open $w $v"
   if {[info exists Tree($w:$v:open)] && $Tree($w:$v:open)==0
      && [info exists Tree($w:$v:children)]
      && [string length $Tree($w:$v:children)]>0} {
      set Tree($w:$v:open) 1
      # puts "Tree:build $v"
      Tree:build $w
      ;# Recurse to a certain depth
      if {$recurse >= [llength [file split $v]]} {
         foreach c $Tree($w:$v:children) {
            Tree:open $w [file join $v $c] $recurse
         } ;# end foreach
      } ;# end if
   }
}
proc Tree:close {w v {recurse 0}} {
   global Tree
   if {[info exists Tree($w:$v:open)] && $Tree($w:$v:open)==1} {
      set Tree($w:$v:open) 0
      Tree:build $w
      if {$recurse} {
         foreach c $Tree($w:$v:children) {
            Tree:close $w [file join $v $c] $recurse
         } ;# end foreach
      }
   }
}
# Internal use only.
# Change highlight to dashed border when focus changes
proc Tree:FocusChange {w In} {
   global Tree
   if {$In} {
      $w itemconfigure $Tree($w:selidx) -state normal -outline {}
   } else {
      $w itemconfigure $Tree($w:selidx) -state disabled -outline black
   }
}
# Draw the selection highlight
proc Tree:drawselection {w} {
   global Tree
   if {[string length $Tree($w:selidx)]} {
      $w delete $Tree($w:selidx)
   }
   set v $Tree($w:selection)
   if {[string length $v]==0} return
   if {![info exists Tree($w:$v:tag)]} return
   set bbox [$w bbox $Tree($w:$v:tag)]
   set Tree($w:activeTag) $Tree($w:$v:tag)
   if {[llength $bbox]==4} {
      set i [eval $w create rectangle $bbox \
         -outline $::Tree(fg) \
         -disabledoutline $::Tree(fg)\
         -fill $::Tree(highlight) \
         -disabledfill $::Tree(bg) \
         -dash 1]
      set Tree($w:selidx) $i
      $w lower $i
   } else {
      set Tree($w:selidx) {}
   }
   ;##
   ;# Change the open/close folder tag
   ;#
   ;# Hack: Folder icon tag is two below text tag (1 below is checkbox)
   catch {
      set tag [expr {$Tree($w:$v:tag)-2}]
      set folder $Tree($w:tag:$tag)
      set icon $Tree($w:$folder:icon)
      ;##
      ;# Configure the icon as open or closed as appropriate
      ;#
      if {[info exists Tree($w:lastFI)]} {
         $w itemconfigure $Tree($w:lastFI) -image $Tree($w:lastIcon).close
      } ;# end if
      $w itemconfigure $tag -image $icon.open
      set Tree($w:lastFI) $tag
      set Tree($w:lastIcon) $icon
   }
}

# Internal use only
# Call Tree:build then next time we're idle
proc Tree:buildwhenidle {w} {
   global Tree
   if {![info exists Tree($w:buildpending)]} {
      set Tree($w:buildpending) 1
      after idle [list Tree:build $w]
   }
}

#
# Return the tag for widget $w that is located
# at real coordinates $x, $y
#
proc Tree:tagat {w x y} {
   set x [$w canvasx $x]
   set y [$w canvasy $y]
   global Tree
   foreach m [$w find overlapping $x $y $x $y] {
   #$Tree($w:$v:tag)
      if {[info exists Tree($w:tag:$m)]} {
         return $m
      }
   }
   return ""
}
#
# Return the full pathname of the label for widget $w that is located
# at real coordinates $x, $y
#
proc Tree:labelat {w x y} {
   set x [$w canvasx $x]
   set y [$w canvasy $y]
   global Tree
   foreach m [$w find overlapping $x $y $x $y] {
      if {[info exists Tree($w:tag:$m)]} {
         return $Tree($w:tag:$m)
      }
   }
   return ""
}
;##
;# Debugging stuff
;#
proc TraceTree {} {
   trace variable ::Tree w DumpTree
}
proc DumpTree {_G I H} {
   upvar $_G G
   puts "$I = $G($I)"
}
# Tree:folderTree teFolders
package provide saTree 1.0

